use super::*;
use smallmap::Map;

#[derive(Debug)]
pub struct Bytes;

type BytesUnit = u128;
type IntermediateUnit = f64;

const KB: BytesUnit = 1024;

lazy_static! {
    static ref SUFFIX: Map<char, BytesUnit> = smallmap::smallmap! {
	{'b' => 1},
	
	{'k' => KB},
	{'m' => KB * KB},
	{'g' => KB * KB * KB},
	{'p' => KB * KB * KB * KB},
    };
}

    #[cfg(feature="f128-math")]
fn multiply(o: IntermediateUnit, u: BytesUnit) -> BytesUnit
{
    (o as f128 * (u as f128)).ceil() as BytesUnit
}

    #[cfg(not(feature="f128-math"))]
fn multiply(o: IntermediateUnit, u: BytesUnit) -> BytesUnit
{
    (o * (u as f64)).ceil() as BytesUnit
}

impl Conversion for Bytes
{
    const UNIT: &'static str = "bytes";

    type Output = BytesUnit;
    type Unit = BytesUnit;

    fn convert(&self, input: &str) -> Result<(Self::Output, Option<Self::Unit>), Self::Error>
    {
	//TODO: Apparently, we cannot parse `f128` from string? See if there's another better way of doing this, or if there are fractional types at all...
	macro_rules! parse {
	    ($non:ident: $type:ty) => (($non).parse::<$type>().map_err(|e| ConversionError::new_for(self, format!("Failed to parse {}: {}", std::any::type_name::<$type>(), e)))?);
	    ($non:ident) => (parse!($non: Self::Output));
	}

	let (o, u) = match input.char_indices().last().map(|(idx, chr)| (chr.to_lowercase().next().unwrap(), &input[..idx])) {
	    Some((chr, _)) if chr.is_numeric() => return Ok((parse!(input: Self::Output), Some(1))),
	    
	    Some((ref chr, non)) if SUFFIX.contains_key(chr) => (parse!(non: IntermediateUnit), *SUFFIX.get(chr).unwrap()),
	    
	    Some((chr, _)) => return Err(ConversionError::new_for(self, format!("Unknown suffix {}", chr))),
	    None => (0 as IntermediateUnit, 1),
	};

	if o.is_sign_negative() {
	    return Err(ConversionError::new_for(self, "Negative byte offsets not supported"));
	}
	
	Ok((multiply(o, u), Some(u)))
    }
}
